from datetime import datetime
import os
from loguru import logger
from rttt.event import Event, EventType
from rttt.connectors.base import Connector
from rttt.connectors.middleware import AsyncMiddleware


class FileLogMiddleware(AsyncMiddleware):

    lut = {
        EventType.LOG: ' # ',
        EventType.OUT: ' > ',
        EventType.IN: ' < ',
        EventType.FLASH: ' ! ',
    }

    def __init__(self, connector: Connector, file_path: str, text: str = '') -> None:
        super().__init__(connector)
        self.open_text = text
        logger.info(f'file_path: {file_path}')
        d = os.path.dirname(file_path)
        if d:
            os.makedirs(d, exist_ok=True)
        self.fd = None
        try:
            self.fd = open(file_path, 'a', encoding='utf-8')
        except Exception as e:
            logger.error(f"Failed to open log file {file_path}: {e}")

    def open(self):
        if self.fd:
            try:
                self.fd.write(f'{"*" * 80}\n')
                center_text = f'{self.open_text:^74}'
                self.fd.write(f'***{center_text}***\n')
                self.fd.write(f'{"*" * 80}\n')
                self.fd.flush()
            except Exception as e:
                logger.error(f"Failed to write header to log file: {e}")
        super().open()

    def close(self):
        super().close()
        if self.fd:
            try:
                self.fd.close()
            except Exception as e:
                logger.error(f"Failed to close log file: {e}")
            self.fd = None

    async def _process(self, event: Event):
        prefix = self.lut.get(event.type, None)
        if prefix and self.fd:
            t = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:23]
            if event.type == EventType.FLASH:
                data = event.data
                status = data.get("status", "")
                if status == "progress":
                    text = f'FLASH [{data.get("percentage", 0):3d}%] {data.get("action", "")}'
                else:
                    text = f'FLASH {status}: {data}'
            else:
                text = event.data
            line = f'{t}{prefix}{text}\n'
            await self._loop.run_in_executor(None, self._write_line, line)

    def _write_line(self, line):
        try:
            self.fd.write(line)
            self.fd.flush()
        except Exception as e:
            logger.error(f"Failed to write to log file: {e}")



from abc import ABCMeta, abstractmethod
from typing import Callable
from loguru import logger
from rttt.event import Event


class Connector(metaclass=ABCMeta):

    def __init__(self):
        self._handler = None

    def on(self, handler: Callable[[Event], None]):
        if self._handler is not None:
            raise RuntimeError(f"Handler already set on {self.__class__.__name__}, call off() first")
        self._handler = handler

    def off(self):
        self._handler = None

    def _emit(self, event: Event):
        if self._handler:
            try:
                self._handler(event)
            except Exception as e:
                logger.error(f"Error in handler: {e}")

    @abstractmethod
    def open(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def handle(self, event: Event):
        pass



from prompt_toolkit.widgets import TextArea, SearchToolbar, Frame, HorizontalLine, ProgressBar, Dialog, Box, Label
from prompt_toolkit.layout.containers import HSplit, VSplit, Window, WindowAlign, ConditionalContainer, FloatContainer, Float
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.margins import NumberedMargin, ScrollbarMargin
from prompt_toolkit.layout.dimension import LayoutDimension
from prompt_toolkit.history import FileHistory
from prompt_toolkit.layout.layout import Layout
from datetime import datetime
from prompt_toolkit.filters import Condition
from prompt_toolkit.application import get_app
from rttt.lexer import LogLexer


class State:

    SHOW_ALL = 0
    SHOW_TERMINAL = 1
    SHOW_LOGGER = 2

    def __init__(self):
        self.show_status_bar = True
        self.scroll_to_end = True
        self.show = self.SHOW_ALL
        self.app = None
        self.flash_visible = False
        self.flash_file = ""
        self.flash_action = ""
        self.flash_error = ""
        self.flash_bar = None

    def is_show_status_bar(self):
        return self.show_status_bar

    def is_show_terminal(self):
        return self.show == self.SHOW_TERMINAL

    def is_show_logger(self):
        return self.show == self.SHOW_LOGGER

    def is_show_all(self):
        return self.show == self.SHOW_ALL

    def show_terminal(self):
        self.show = self.SHOW_TERMINAL

    def show_logger(self):
        self.show = self.SHOW_LOGGER

    def show_all(self):
        self.show = self.SHOW_ALL

    def scroll_to_end_toggle(self):
        self.scroll_to_end = not self.scroll_to_end
        return self.scroll_to_end

    def has_focus(self, value):
        return self.app.layout.has_focus(value) if self.app else False

    def set_app(self, app):
        self.app = app


def create_terminal_window():
    """
    Create the interactive terminal window.
    """
    terminal_search = SearchToolbar(ignore_case=True, vi_mode=True)
    terminal_window = TextArea(
        text="",
        scrollbar=True,
        line_numbers=True,
        focusable=True,
        focus_on_click=True,
        read_only=True,
        search_field=terminal_search,
    )
    return terminal_window, terminal_search


def create_terminal_input(state: State, history_file=None):
    """
    Create the terminal input field.
    """
    input_history = FileHistory(history_file) if history_file else None
    input_search = SearchToolbar(ignore_case=True)
    input_field = TextArea(
        height=1,
        prompt=lambda: [('class:cyan', 'Command: ')] if state.has_focus(input_field) else 'Command: ',
        style="class:input-field",
        multiline=False,
        wrap_lines=False,
        search_field=input_search,
        history=input_history,
        focusable=True,
        focus_on_click=True,
    )
    return input_field, input_search


def create_logger_window():
    """
    Create the logger window.
    """
    logger_search = SearchToolbar(ignore_case=True, vi_mode=True)
    logger_window = TextArea(
        scrollbar=True,
        line_numbers=True,
        focusable=True,
        focus_on_click=True,
        read_only=True,
        search_field=logger_search,
        lexer=LogLexer(),
    )
    return logger_window, logger_search


def create_status_bar(state):
    """
    Create the status bar for the console.
    """
    def get_statusbar_text():
        return [
            ('class:title', ' HARDWARIO RTTT Console     '),
            ('class:title', ' <F3> Focus '),
            ('class:title', ' <F5> Pause ') if state.scroll_to_end else ('class:yellow', ' <F5> Pause '),
            ('class:title', ' <F8> Clear '),
            ('class:title', ' <F10> Exit (or Ctrl-<F10>) '),
            ('class:title', ' [Shift-]<Tab> Cycle '),
        ]

    def get_statusbar_time():
        return datetime.now().strftime('%b %d, %Y  %H:%M:%S')

    return ConditionalContainer(
        content=VSplit([
            Window(
                FormattedTextControl(get_statusbar_text), style="class:status"
            ),
            Window(
                FormattedTextControl(get_statusbar_time),
                style="class:status.right",
                width=24,
                align=WindowAlign.RIGHT,
            ),
        ],
            height=LayoutDimension.exact(1),
            style="class:statusbar",),
        filter=Condition(state.is_show_status_bar)
    )


def create_layout(state, history_file):
    """
    Create the layout
    """
    input_field, input_search = create_terminal_input(state, history_file)
    terminal_window, terminal_search = create_terminal_window()
    logger_window, logger_search = create_logger_window()

    status_bar = create_status_bar(state)

    hs_terminal = HSplit(
        [
            terminal_window,
            terminal_search,
            HorizontalLine(),
            input_field,
            input_search,
        ]
    )

    hs_logger = HSplit([
        logger_window,
        logger_search
    ])

    flash_bar = ProgressBar()
    state.flash_bar = flash_bar

    flash_overlay = Float(
        content=ConditionalContainer(
            content=Box(
                body=Frame(
                    body=HSplit([
                        Window(FormattedTextControl(lambda: state.flash_file), height=1, align=WindowAlign.CENTER),
                        Window(height=1),
                        ConditionalContainer(
                            content=HSplit([
                                Window(FormattedTextControl(lambda: state.flash_action), height=1, align=WindowAlign.CENTER),
                                flash_bar,
                            ]),
                            filter=Condition(lambda: not state.flash_error),
                        ),
                        ConditionalContainer(
                            content=Window(FormattedTextControl(lambda: state.flash_error), height=1, align=WindowAlign.CENTER, style="fg:#ff4444 bold"),
                            filter=Condition(lambda: bool(state.flash_error)),
                        ),
                    ]),
                    title="Flash",
                ),
                style="bg:#222222 fg:#eeeeee",
            ),
            filter=Condition(lambda: state.flash_visible),
        ),
    )

    root_container = FloatContainer(
        content=HSplit(
            [
                ConditionalContainer(
                    content=VSplit(
                        [
                            Frame(hs_terminal, title="Interactive Terminal"),
                            Frame(hs_logger, title="Device Log")
                        ]
                    ),
                    filter=Condition(state.is_show_all)),
                ConditionalContainer(
                    content=hs_terminal,
                    filter=Condition(state.is_show_terminal)
                ),
                ConditionalContainer(
                    content=hs_logger,
                    filter=Condition(state.is_show_logger)
                ),
                status_bar
            ]
        ),
        floats=[flash_overlay],
        style="bg:#111111 fg:#eeeeee",
    )

    return root_container, input_field, terminal_window, logger_window



import re
from typing import Callable
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.document import Document
from prompt_toolkit.styles.named_colors import NAMED_COLORS
from prompt_toolkit.formatted_text.base import StyleAndTextTuples


DEF_LOG_LEXER_COLOR_TABLE = {
    'default': '#eeeeee',
    'X': NAMED_COLORS['Blue'],
    'D': NAMED_COLORS['Magenta'],
    'I': NAMED_COLORS['Green'],
    'W': NAMED_COLORS['Yellow'],
    'E': NAMED_COLORS['Red'],
    'dbg': NAMED_COLORS['Magenta'],
    'inf': NAMED_COLORS['Green'],
    'wrn': NAMED_COLORS['Yellow'],
    'err': NAMED_COLORS['Red'],
}

LOG_PATERN = r'^(\[.*?\].*?<(\w+)\>)(.*)'
OLD_LOG_PATERN = r'^(#.*?\d(?:\.\d+)? <(\w)\>)(.*)'


class LogLexer(Lexer):
    def __init__(self, colors: dict = DEF_LOG_LEXER_COLOR_TABLE) -> None:
        super().__init__()
        self.re = re.compile(LOG_PATERN)
        self.re_old = re.compile(OLD_LOG_PATERN)
        self.colors = colors or DEF_LOG_LEXER_COLOR_TABLE
        self.def_color = self.colors.get('default', '#eeeeee')

    def lex_document(self, document: Document) -> Callable[[int], StyleAndTextTuples]:
        lines = document.lines

        def get_line(lineno: int) -> StyleAndTextTuples:
            line = lines[lineno]
            g = self.re.match(line)
            if not g:
                g = self.re_old.match(line)
            if g:
                color = self.colors.get(g.group(2), self.def_color)
                return [(color, g.group(1)), (self.def_color, g.group(3))]

            return [(self.def_color, line)]

        return get_line



from rttt.connectors.base import Connector
from rttt.connectors.middleware import Middleware, AsyncMiddleware, BufferedMiddleware
from rttt.connectors.demo import DemoConnector
from rttt.connectors.file_log import FileLogMiddleware
from rttt.connectors.mcp_server import MCPMiddleware
from rttt.connectors.pylink_rtt import PyLinkRTTConnector

# Backward compatibility alias (deprecated, use FileLogMiddleware)
FileLogConnector = FileLogMiddleware



import asyncio
from abc import abstractmethod
from loguru import logger
from rttt.connectors.base import Connector
from rttt.event import Event
from rttt.utils import ensure_loop, shutdown_loop


class Middleware(Connector):

    def __init__(self, connector: Connector):
        super().__init__()
        self.connector = connector
        self.connector.on(self._on)

    def open(self):
        self.connector.open()

    def close(self):
        self.connector.close()

    def handle(self, event: Event):
        self.connector.handle(event)

    def _on(self, event: Event):
        self._emit(event)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.connector!r})"


class AsyncMiddleware(Middleware):
    """Middleware with a background asyncio loop for async processing.

    Events are immediately passed downstream via _emit, then enqueued
    for async background processing (e.g. file logging).

    Override `_process(event)` for background work.
    """

    def __init__(self, connector: Connector):
        super().__init__(connector)
        self._accepting = False
        self._loop = None
        self._queue = None
        self._thread = None
        self._task = None

    def open(self):
        self._loop, self._thread = ensure_loop(
            name=f"RTTT-{self.__class__.__name__}-{id(self)}"
        )
        self._queue = asyncio.Queue()
        self._task = self._loop.create_task(self._worker())
        self._accepting = True
        super().open()

    def close(self):
        self._accepting = False
        super().close()
        if self._loop and self._loop.is_running():
            async def _shutdown():
                await self._queue.join()
                self._task.cancel()
            future = asyncio.run_coroutine_threadsafe(_shutdown(), self._loop)
            try:
                future.result(timeout=5)
            except Exception:
                pass
        shutdown_loop(self._loop, self._thread)
        self._thread = None
        self._loop = None

    def _on(self, event: Event):
        self._emit(event)
        if self._accepting:
            self._loop.call_soon_threadsafe(self._queue.put_nowait, event)

    async def _worker(self):
        while True:
            try:
                event = await self._queue.get()
                try:
                    await self._process(event)
                except Exception as e:
                    logger.error(f"Error in {self.__class__.__name__}._process: {e}")
                self._queue.task_done()
            except asyncio.CancelledError:
                break

    @abstractmethod
    async def _process(self, event: Event):
        pass


class BufferedMiddleware(AsyncMiddleware):
    """Middleware that buffers events in a queue before passing downstream.

    Events are NOT emitted immediately - they are processed first.
    Override `_process(event)` to transform/filter events.
    Return the event to emit it, or None to drop it.
    """

    def _on(self, event: Event):
        if self._accepting:
            self._loop.call_soon_threadsafe(self._queue.put_nowait, event)

    async def _worker(self):
        while True:
            try:
                event = await self._queue.get()
                try:
                    result = await self._process(event)
                    if result is not None:
                        self._emit(result)
                except Exception as e:
                    logger.error(f"Error in {self.__class__.__name__}._process: {e}")
                self._queue.task_done()
            except asyncio.CancelledError:
                break

    @abstractmethod
    async def _process(self, event: Event) -> Event | None:
        pass



import os
import pylink
import time
import threading
from loguru import logger
from rttt.connectors.base import Connector
from rttt.event import Event, EventType


class PyLinkRTTConnector(Connector):

    def __init__(self, jlink: pylink.JLink, terminal_buffer=0, logger_buffer=1, latency=50, block_address=None,) -> None:
        super().__init__()
        self.jlink = jlink
        self.block_address = block_address
        self.rtt_read_delay = latency / 1000.0
        self.is_running = False
        self.thread = None
        self.terminal_buffer = terminal_buffer
        self.terminal_buffer_up_size = 0
        self.terminal_buffer_down_size = 0
        self.logger_buffer = logger_buffer
        self.log_up_size = 0
        self._flash_lock = threading.Lock()

    def start(self):
        """Start RTT and the read thread."""
        self._cache = {0: '', 1: ''}

        logger.info(f"Starting RTT{' control block found at 0x{:08X}'.format(self.block_address) if self.block_address else ''}")
        self.jlink.rtt_start(self.block_address)

        for _ in range(100):
            try:
                num_up = self.jlink.rtt_get_num_up_buffers()
                num_down = self.jlink.rtt_get_num_down_buffers()
                logger.info(f'RTT started, {num_up} up bufs, {num_down} down bufs.')
                break
            except pylink.errors.JLinkRTTException:
                time.sleep(0.1)
        else:
            raise Exception('Failed to find RTT block')

        if num_up == 0:
            raise Exception('No RTT up buffers found')

        if num_up < self.terminal_buffer:
            raise Exception(f'Shell buffer UP {self.terminal_buffer} not found')

        if num_up < self.logger_buffer:
            raise Exception(f'Log buffer UP {self.logger_buffer} not found')

        if num_down < self.terminal_buffer:
            raise Exception(f'Shell buffer DOWN {self.terminal_buffer} not found')

        self.is_running = True

        for i in range(num_up):
            desc = self.jlink.rtt_get_buf_descriptor(i, 1)
            try:
                name = desc.name
            except UnicodeDecodeError:
                name = desc.acName.decode('utf-8', errors='replace')
            logger.info(f'Up buffer {i}: {name} <Index={desc.BufferIndex}, Size={desc.SizeOfBuffer}>')
            if i == self.terminal_buffer:
                self.terminal_buffer_up_size = desc.SizeOfBuffer
            elif i == self.logger_buffer:
                self.log_up_size = desc.SizeOfBuffer
        for i in range(num_down):
            desc = self.jlink.rtt_get_buf_descriptor(i, 0)
            try:
                name = desc.name
            except UnicodeDecodeError:
                name = desc.acName.decode('utf-8', errors='replace')
            logger.info(f'Down buffer {i}: {name} <Index={desc.BufferIndex}, Size={desc.SizeOfBuffer}>')
            if i == self.terminal_buffer:
                self.terminal_buffer_down_size = desc.SizeOfBuffer

        self.thread = threading.Thread(target=self._read_task, daemon=True)
        self.thread.start()

    def stop(self):
        """Stop the read thread and RTT."""
        if not self.is_running:
            return
        self.is_running = False
        if self.thread:
            self.thread.join()
            self.thread = None
        self.jlink.rtt_stop()

    def open(self):
        super().open()
        self.start()
        self._emit(Event(EventType.OPEN, ''))
        logger.info('RTT opened')

    def close(self):
        super().close()
        logger.info('Closing RTT')
        self.stop()
        self._emit(Event(EventType.CLOSE, ''))
        logger.info('RTT closed')

    def handle(self, event: Event):
        logger.info(f'handle: {event.type} {event.data}')
        if event.type == EventType.IN:
            logger.info(f'RTT write shell buffer {self.terminal_buffer} buffer size {self.terminal_buffer_down_size}')
            if not self.terminal_buffer_down_size:
                raise Exception(f'Shell buffer DOWN {self.terminal_buffer} has zero size')
            data = bytearray(f'{event.data}\n', "utf-8")
            for i in range(0, len(data), self.terminal_buffer_down_size):
                chunk = data[i:i + self.terminal_buffer_down_size]
                self.jlink.rtt_write(self.terminal_buffer, list(chunk))
        elif event.type == EventType.FLASH:
            self.flash(event.data.get('file'), event.data.get('addr', 0))
            return
        self._emit(event)

    def flash(self, file_path: str, addr: int = 0):
        """Flash firmware file to device. Stops RTT, flashes, restarts RTT."""
        if not self._flash_lock.acquire(blocking=False):
            self._emit(Event(EventType.FLASH, {
                "status": "error", "file": file_path,
                "error": "Flash operation already in progress"
            }))
            return

        try:
            if not os.path.isfile(file_path):
                self._emit(Event(EventType.FLASH, {
                    "status": "error", "file": file_path,
                    "error": f"File not found: {file_path}"
                }))
                return

            ext = os.path.splitext(file_path)[1].lower()
            if ext not in ('.hex', '.bin', '.elf', '.srec'):
                self._emit(Event(EventType.FLASH, {
                    "status": "error", "file": file_path,
                    "error": f"Unsupported file format: {ext}"
                }))
                return

            self._emit(Event(EventType.FLASH, {"status": "start", "file": file_path}))

            was_running = self.is_running
            if was_running:
                self.stop()

            def on_progress(action, progress_string, percentage):
                if isinstance(action, bytes):
                    action = action.decode('utf-8', errors='replace')
                if isinstance(progress_string, bytes):
                    progress_string = progress_string.decode('utf-8', errors='replace')
                if action == "Erase":
                    return
                self._emit(Event(EventType.FLASH, {
                    "status": "progress", "action": action,
                    "message": progress_string,
                    "percentage": min(percentage, 100),
                }))

            try:
                logger.info(f'Flash: flashing {file_path} at 0x{addr:08X}')
                bytes_flashed = self.jlink.flash_file(file_path, addr, on_progress=on_progress)
                logger.info(f'Flash: complete, {bytes_flashed} bytes')
                self.jlink.reset(ms=10, halt=False)
                self._emit(Event(EventType.FLASH, {
                    "status": "done", "file": file_path, "bytes_flashed": bytes_flashed
                }))
            except Exception as e:
                logger.error(f'Flash error: {e}')
                self._emit(Event(EventType.FLASH, {
                    "status": "error", "file": file_path, "error": str(e)
                }))

            if was_running:
                try:
                    self.start()
                except Exception as e:
                    logger.error(f'Flash: failed to restart RTT: {e}')
                    self._emit(Event(EventType.FLASH, {
                        "status": "error", "file": file_path,
                        "error": f"RTT restart failed: {e}"
                    }))
        finally:
            self._flash_lock.release()

    def _read_task(self):
        while self.is_running:
            channels = [
                (self.terminal_buffer, min(1000, self.terminal_buffer_up_size), EventType.OUT),
                (self.logger_buffer, min(1000, self.log_up_size), EventType.LOG)
            ]
            for idx, num_bytes, event_type in channels:
                if idx is None:
                    continue
                try:
                    data = self.jlink.rtt_read(idx, num_bytes)
                    if data:
                        lines = bytes(data).decode('utf-8', errors="backslashreplace")
                        if lines:
                            lines = self._cache[idx] + lines

                            while True:
                                end = lines.find('\n')
                                if end < 0:
                                    self._cache[idx] = lines
                                    break

                                line = lines[:end]
                                lines = lines[end + 1:]

                                if line.endswith('\r'):
                                    line = line[:-1]

                                self._emit(Event(event_type, line))
                except Exception as e:
                    logger.error(f'Error reading RTT buffer {idx}: {e}')
            time.sleep(self.rtt_read_delay)



import asyncio
import os
from collections import deque
from loguru import logger
from mcp.server.fastmcp import FastMCP
from rttt.connectors.base import Connector
from rttt.connectors.middleware import AsyncMiddleware
from rttt.event import Event, EventType
from rttt.utils import parse_listen


class MCPMiddleware(AsyncMiddleware):
    """Middleware that runs an MCP server for AI tool integration.

    Collects events into ring buffers and exposes MCP tools for
    reading terminal/log output and sending commands to the device.
    Uses Streamable HTTP transport.
    """

    def __init__(self, connector: Connector, listen: str = "127.0.0.1:8090", max_lines: int = 5000, name: str = "Device Console"):
        super().__init__(connector)
        self.host, self.port = parse_listen(listen)
        self.max_lines = max_lines
        self._terminal_lines = deque(maxlen=max_lines)
        self._log_lines = deque(maxlen=max_lines)
        self._terminal_cursor = 0
        self._log_cursor = 0
        self._terminal_event = asyncio.Event()
        self._flash_events = []
        self._flash_done = None
        self._mcp = FastMCP(name, host=self.host, port=self.port, log_level="CRITICAL", stateless_http=True)
        self._register_tools()
        self._server_task = None

    def open(self):
        super().open()
        self._server_task = asyncio.run_coroutine_threadsafe(
            self._mcp.run_streamable_http_async(), self._loop
        )
        logger.info(f'MCP server starting on http://{self.host}:{self.port}/mcp')

    def close(self):
        if self._server_task:
            self._server_task.cancel()
            try:
                self._server_task.result(timeout=3)
            except Exception:
                pass
        super().close()

    async def _process(self, event: Event):
        if event.type in (EventType.OUT, EventType.IN):
            direction = "in" if event.type == EventType.IN else "out"
            self._terminal_lines.append({"direction": direction, "text": event.data})
            self._terminal_cursor += 1
            self._terminal_event.set()
        elif event.type == EventType.LOG:
            self._log_lines.append(event.data)
            self._log_cursor += 1
        elif event.type == EventType.FLASH:
            self._flash_events.append(event.data)
            if event.data.get("status") in ("done", "error") and self._flash_done is not None:
                self._flash_done.set()

    def _register_tools(self):
        middleware = self

        @self._mcp.tool()
        async def send_command(command: str, timeout: float = 2.0) -> dict:
            """Send a shell command to the connected device.

            Waits up to `timeout` seconds for the device to finish responding,
            then returns all terminal output produced after the command along
            with a `log_cursor` that can be passed to `read_log` to fetch only
            the logs that appeared since this command was sent.
            """
            try:
                log_cursor = middleware._log_cursor
                terminal_cursor = middleware._terminal_cursor
                middleware._terminal_event.clear()
                middleware.connector.handle(Event(EventType.IN, command))

                # Wait for output lines to arrive; stop after `timeout` seconds
                # of silence (no new terminal data).
                deadline = asyncio.get_event_loop().time() + timeout
                while True:
                    remaining = deadline - asyncio.get_event_loop().time()
                    if remaining <= 0:
                        break
                    middleware._terminal_event.clear()
                    try:
                        await asyncio.wait_for(middleware._terminal_event.wait(), timeout=remaining)
                        # Got new data — reset the silence deadline
                        deadline = asyncio.get_event_loop().time() + 0.3
                    except asyncio.TimeoutError:
                        break

                # Collect output lines that appeared after the command
                # Use cursor difference to slice from the right end of the deque,
                # which is safe even when the deque has wrapped around.
                new_count = middleware._terminal_cursor - terminal_cursor
                all_lines = list(middleware._terminal_lines)
                output = all_lines[-new_count:] if new_count > 0 else []
                # Filter out the echo of the sent command
                output = [line for line in output if line["direction"] != "in"]
                return {
                    "status": "ok",
                    "command": command,
                    "output": output,
                    "log_cursor": log_cursor,
                    "log_cursor_end": middleware._log_cursor,
                }
            except Exception as e:
                return {
                    "status": "error",
                    "command": command,
                    "error": str(e),
                }

        @self._mcp.tool()
        def read_terminal(lines: int = 100) -> list:
            """Read recent terminal output. Returns JSON array with direction (in/out) and text."""
            return list(middleware._terminal_lines)[-lines:]

        @self._mcp.tool()
        def read_log(lines: int = -1, after_cursor: int = -1, pattern: str = "") -> list:
            """Read log output from the device. Logs are kept in a ring buffer
            and can be read repeatedly. If you need buffer state, use the `status` tool.

            Args:
                lines: Number of lines to return (-1 = all buffered).
                after_cursor: Return only lines after this cursor (from `send_command`).
                pattern: Regex filter (case-insensitive). Only matching lines are returned.
            """
            import re
            all_logs = list(middleware._log_lines)
            if after_cursor >= 0:
                total = middleware._log_cursor
                available = len(all_logs)
                skip = after_cursor - (total - available)
                if skip < 0:
                    skip = 0
                result = all_logs[skip:]
            elif lines >= 0:
                result = all_logs[-lines:] if lines > 0 else []
            else:
                result = all_logs
            if pattern:
                try:
                    regex = re.compile(pattern, re.IGNORECASE)
                    result = [line for line in result if regex.search(line)]
                except re.error:
                    pass
            return result

        @self._mcp.tool()
        def status() -> dict:
            """Get session statistics: total and buffered terminal/log line counts and current cursors."""
            return {
                "terminal_total": middleware._terminal_cursor,
                "terminal_buffered": len(middleware._terminal_lines),
                "log_total": middleware._log_cursor,
                "log_buffered": len(middleware._log_lines),
                "buffer_size": middleware.max_lines,
            }

        @self._mcp.tool()
        async def flash(file_path: str, addr: int = 0) -> dict:
            """Flash a firmware file (.hex, .bin, .elf, .srec) to the target device.

            Stops the connection, programs the flash memory, resets the device,
            and restarts the connection. Progress is reported to the console during flashing.

            Common firmware locations:
            - Zephyr/west: build/merged.hex (or build/<app>/merged.hex)
            - nRF Connect SDK: build/merged.hex
            - Generic: build/firmware.hex, build/output.bin

            Args:
                file_path: Absolute path to the firmware file.
                addr: Start address for flashing (default 0, ignored for .hex files).
            """
            if not os.path.isfile(file_path):
                return {"status": "error", "error": f"File not found: {file_path}"}

            middleware._flash_events = []
            middleware._flash_done = asyncio.Event()

            try:
                loop = asyncio.get_event_loop()
                await loop.run_in_executor(
                    None,
                    middleware.connector.handle,
                    Event(EventType.FLASH, {"file": file_path, "addr": addr})
                )

                try:
                    await asyncio.wait_for(middleware._flash_done.wait(), timeout=120.0)
                except asyncio.TimeoutError:
                    return {"status": "error", "error": "Flash operation timed out"}

                final = middleware._flash_events[-1] if middleware._flash_events else {}
                result = {
                    "status": final.get("status", "unknown"),
                    "file": file_path,
                    "events": middleware._flash_events,
                }
                if "bytes_flashed" in final:
                    result["bytes_flashed"] = final["bytes_flashed"]
                if "error" in final:
                    result["error"] = final["error"]
                return result

            except Exception as e:
                return {"status": "error", "error": str(e)}
            finally:
                middleware._flash_done = None



import os
import asyncio
from loguru import logger
from prompt_toolkit.application import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.styles import Style, Priority
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.document import Document
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.filters import Condition
from rttt.ui import State, create_layout
from rttt.utils import truncate_path
from rttt.connectors.base import Connector
from rttt.event import Event, EventType


class Console:

    def __init__(self, connector: Connector, history_file=None):
        self.connector = connector
        self.state = State()
        self.exception = None

        if history_file:
            d = os.path.dirname(history_file)
            if d:
                os.makedirs(d, exist_ok=True)

        root_container, input_field, terminal_window, logger_window = create_layout(self.state, history_file)
        self.input_field = input_field
        self.input_field.accept_handler = self._input_accept_handler

        bindings = KeyBindings()

        @bindings.add("c-insert", eager=True)  # TODO: check
        @bindings.add("c-c", eager=True)
        def _(event):
            data = None
            if self.has_focus(terminal_window):
                data = terminal_window.buffer.copy_selection()
            elif self.has_focus(logger_window):
                data = logger_window.buffer.copy_selection()
            else:
                return
            if data.text:
                try:
                    event.app.clipboard.set_data(data)
                except Exception as e:
                    logger.error(e)

        @bindings.add("f5", eager=True)
        def _(event):
            if self.state.scroll_to_end_toggle():
                self.terminal_buffer.cursor_position = len(self.terminal_buffer.text)
                self.logger_buffer.cursor_position = len(
                    self.logger_buffer.text)

        @bindings.add("f8", eager=True)
        def _(event):
            self.terminal_buffer.set_document(Document(''), True)
            self.logger_buffer.set_document(Document(''), True)

        @bindings.add("f3", eager=True)
        def _(event):
            if not self.state.is_show_all():
                self.state.show_all()
            elif self.has_focus(self.input_field) or self.has_focus(terminal_window):
                self.state.show_terminal()
            elif self.has_focus(logger_window):
                self.state.show_logger()

        @bindings.add("c-q", eager=True)
        @bindings.add("f10", eager=True)
        @bindings.add("c-f10", eager=True)
        def _(event):
            event.app.exit()

        bindings.add("tab")(focus_previous)
        bindings.add("s-tab")(focus_next)

        self.terminal_buffer = terminal_window.buffer
        self.logger_buffer = logger_window.buffer

        self.app = Application(
            layout=Layout(root_container, focused_element=self.input_field),
            key_bindings=bindings,
            mouse_support=Condition(lambda: not self.state.is_show_all()),
            full_screen=True,
            refresh_interval=1,
            enable_page_navigation_bindings=True,
            clipboard=PyperclipClipboard(),
            style=Style.from_dict({
                'border': '#888888',
                'message': 'bg:#bbee88 #222222',
                'statusbar': 'noreverse bg:gray #000000',
                'progress-bar.used': 'bg:#4488cc',
            }, priority=Priority.MOST_PRECISE)
        )

        self.state.set_app(self.app)

    def run(self):
        async def event_task():
            with logger.catch(message='event_task', reraise=True):
                while True:
                    event = await self.events.get()
                    logger.debug(f'event: {str(event.type)} {event.data}')
                    if event.type == EventType.LOG:
                        self._buffer_insert_text(self.logger_buffer, f'{event.data}\n')
                    elif event.type == EventType.OUT:
                        self._buffer_insert_text(self.terminal_buffer, f'{event.data}\n')
                    elif event.type == EventType.IN:
                        self._buffer_insert_text(self.terminal_buffer, f'{event.data}\n')
                    elif event.type == EventType.FLASH:
                        data = event.data
                        status = data.get("status", "")
                        if status == "start":
                            self.state.flash_file = truncate_path(data.get("file", ""))
                            self.state.flash_action = ""
                            self.state.flash_error = ""
                            self.state.flash_bar.percentage = 0
                            self.state.flash_visible = True
                        elif status == "progress":
                            self.state.flash_bar.percentage = min(data.get("percentage", 0), 100)
                            self.state.flash_action = data.get("action", "")
                        elif status == "done":
                            self.state.flash_visible = False
                        elif status == "error":
                            self.state.flash_error = data.get("message", "Flash error")
                        self.app.invalidate()

        def pre_run():
            self.events = asyncio.Queue()

            def connector_handle_event(event: Event):
                try:
                    self.events.put_nowait(event)
                except Exception as e:
                    logger.exception(f"Failed to queue event: {event}")

            self.connector.on(connector_handle_event)
            self.app.create_background_task(event_task())
            self.connector.open()

        try:
            self.app.run(pre_run=pre_run)
        finally:
            self.connector.close()

    def exit(self, exception=None):
        self.exception = exception
        self.app.exit()

    def has_focus(self, window):
        return self.app.layout.has_focus(window)

    def _input_accept_handler(self, buff: Buffer) -> bool:
        with logger.catch(message='_input_accept_handler', reraise=True):
            text = f'{buff.text}\n'
            for line in text.splitlines():
                self.connector.handle(Event(EventType.IN, line))
            return False  # false to keep the text in the buffer

    def _buffer_insert_text(self, buffer, line):
        changed = buffer._set_text(buffer.text + line)
        if changed:
            if self.state.scroll_to_end:
                buffer.cursor_position = len(buffer.text)
            buffer._text_changed()



import time
import threading
from loguru import logger
from rttt.event import Event, EventType
from rttt.connectors.base import Connector


class DemoConnector(Connector):
    def __init__(self, delay=0.5) -> None:
        super().__init__()
        self.i = 0
        self.delay = delay
        self.is_running = False

    def handle(self, event: Event):
        logger.info(f'handle: {event.type} {event.data}')
        self._emit(event)

    def open(self):
        super().open()
        logger.info('open')
        self.is_running = True
        self.thread = threading.Thread(target=self._task, daemon=True)
        self.thread.start()

    def close(self):
        super().close()
        logger.info('close')
        if not self.is_running:
            return
        self.is_running = False
        self.thread.join()
        self._emit(Event(EventType.CLOSE, ''))

    def _task(self):
        self._emit(Event(EventType.OPEN, ''))
        while self.is_running:
            self.i += 1
            if self.i % 2 == 0:
                self._emit(Event(EventType.LOG, f'log {self.i}'))
            else:
                self._emit(Event(EventType.OUT, f'term {self.i}'))
            time.sleep(self.delay)



import os
import sys
import click
import time
import yaml
import pylink
from loguru import logger
from rttt import __version__ as version
from rttt.connectors import PyLinkRTTConnector, FileLogMiddleware, MCPMiddleware, DemoConnector
from rttt.console import Console

DEFAULT_LOG_FILE = os.path.expanduser("~/.hardwario/rttt.log")
DEFAULT_HISTORY_FILE = os.path.expanduser(f"~/.rttt_history")
DEFAULT_CONSOLE_FILE = os.path.expanduser(f"~/.rttt_console")
DEFAULT_JLINK_SPEED_KHZ = 2000
DEFAULT_MCP_LISTEN = '127.0.0.1:8090'


def get_default_map():
    for cf in ['.rttt.yaml', os.path.expanduser('~/.rttt.yaml'), os.path.expanduser('~/.config/rttt.yaml')]:
        if os.path.exists(cf):
            logger.debug('Loading config from: {}', cf)
            with open(cf, 'r') as f:
                return yaml.safe_load(f)
    return {}


class IntOrHexParamType(click.ParamType):
    name = 'number'

    def convert(self, value, param, ctx):
        try:
            return int(value, 0)
        except ValueError:
            self.fail(f'{value} is not a valid integer or hex value', param, ctx)


@click.command('rttt')
@click.version_option(version, prog_name='rttt')
@click.option('--serial', type=int, metavar='SERIAL_NUMBER', help='J-Link serial number', show_default=True)
@click.option('--device', type=str, metavar='DEVICE', help='J-Link Device name')
@click.option('--speed', type=int, metavar="SPEED", help='J-Link clock speed in kHz', default=DEFAULT_JLINK_SPEED_KHZ, show_default=True)
@click.option('--reset', is_flag=True, help='Reset application firmware.')
@click.option('--address', metavar="ADDRESS", type=IntOrHexParamType(), help='RTT block address.')
@click.option('--terminal-buffer', type=int, help='RTT Terminal buffer index.', show_default=True, default=0)
@click.option('--logger-buffer', type=int, help='RTT Logger buffer index.', show_default=True, default=1)
@click.option('--latency', type=int, help='Latency for RTT readout in ms.', show_default=True, default=50)
@click.option('--history-file', type=click.Path(writable=True), show_default=True, default=DEFAULT_HISTORY_FILE)
@click.option('--console-file', type=click.Path(writable=True), show_default=True, default=DEFAULT_CONSOLE_FILE)
@click.option('--mcp/--no-mcp', is_flag=True, help='Enable MCP server.', show_default=True, default=False)
@click.option('--mcp-listen', type=str, help='MCP server listen address [host:]port.', show_default=True, default=DEFAULT_MCP_LISTEN)
def cli(serial, device, speed, reset, address, terminal_buffer, logger_buffer, latency, history_file, console_file, mcp, mcp_listen):
    '''HARDWARIO Real Time Transfer Terminal Console.'''

    if not device:
        device = click.prompt('Device')

    jlink = pylink.JLink()
    jlink.open(serial_no=serial)
    jlink.set_speed(speed)
    jlink.set_tif(pylink.enums.JLinkInterfaces.SWD)

    logger.info(f'J-Link dll version: {jlink.version}')
    logger.info(f'J-Link dll compile_date: {jlink.compile_date}')
    try:
        logger.info(f'J-Link dll path: {jlink._library._path}')
    except Exception as _:
        pass
    logger.info(f'J-Link serial_number: {jlink.serial_number}')
    logger.info(f'J-Link firmware_version: {jlink.firmware_version}')
    logger.info(f'J-Link firmware_outdated: {jlink.firmware_outdated()}')
    logger.info(f'J-Link firmware_newer: {jlink.firmware_newer()}')

    jlink.connect(device)

    if reset:
        jlink.reset()
        jlink.go()
        time.sleep(1)

    connector = PyLinkRTTConnector(jlink, terminal_buffer, logger_buffer, latency, block_address=address)

    if mcp:
        connector = MCPMiddleware(connector, listen=mcp_listen)

    if console_file:
        text = f'Device: {device} J-Link sn: {serial}' if serial else f'Device: {device}'
        connector = FileLogMiddleware(connector, console_file, text=text)

    console = Console(connector, history_file=history_file)
    console.run()


def main():
    '''Application entry point.'''

    os.makedirs(os.path.expanduser("~/.hardwario"), exist_ok=True)

    logger.remove()
    logger.add(DEFAULT_LOG_FILE,
               format='{time} | {level} | {name}.{function}: {message}',
               level='TRACE',
               rotation='10 MB',
               retention=3)

    logger.debug('Argv: {}', sys.argv)
    logger.debug('Version: {}', version)

    try:
        with logger.catch(reraise=True, exclude=KeyboardInterrupt):
            default_map = get_default_map()
            logger.debug('Loaded config: {}', default_map)
            cli(auto_envvar_prefix='RTTT', default_map=default_map)
    except KeyboardInterrupt:
        pass
    except Exception as e:
        # raise e
        click.secho(str(e), err=True, fg='red')
        if os.getenv('DEBUG', False):
            raise e
        sys.exit(1)



import asyncio
import threading


def ensure_loop(name: str = 'RTTT') -> tuple[asyncio.AbstractEventLoop, threading.Thread | None]:
    """Get or create an asyncio event loop.

    If called from async context, returns the running loop (no thread).
    Otherwise creates a new loop running in a background thread.
    """
    try:
        loop = asyncio.get_running_loop()
        return loop, None
    except RuntimeError:
        loop = asyncio.new_event_loop()
        ready = threading.Event()

        def run():
            asyncio.set_event_loop(loop)
            ready.set()
            loop.run_forever()

        thread = threading.Thread(target=run, daemon=True, name=name)
        thread.start()
        ready.wait(timeout=1.0)
        return loop, thread


def shutdown_loop(loop: asyncio.AbstractEventLoop, thread: threading.Thread | None):
    """Stop and clean up an event loop created by ensure_loop."""
    if thread and loop.is_running():
        loop.call_soon_threadsafe(loop.stop)
        thread.join()
    if not loop.is_closed():
        loop.close()


def parse_listen(listen: str, default_host: str = '127.0.0.1', default_port: int = 8090) -> tuple[str, int]:
    """Parse a listen address string into (host, port).

    Accepts formats: 'host:port', ':port', 'port'.
    """
    if ':' in listen:
        host, port_str = listen.rsplit(':', 1)
        return (host or default_host), int(port_str)
    return default_host, int(listen)


def truncate_path(path, max_length=100):
    """Truncate path to last 4 directory levels if longer than max_length."""
    if len(path) <= max_length:
        return path
    parts = path.replace("\\", "/").split("/")
    # Keep last 4 parts (3 dirs + filename)
    if len(parts) > 4:
        return ".../" + "/".join(parts[-4:])
    return path



import enum


@enum.unique
class EventType(enum.Enum):
    OPEN = 'open'
    CLOSE = 'close'
    OUT = 'out'  # terminal line out
    IN = 'in'    # terminal line in
    LOG = 'log'  # logger line out
    FLASH = 'flash'  # flash programming event


class Event:
    def __init__(self, type: EventType, data):
        self.type = type
        self.data = data
